export type Nullable<T> = T | undefined

export interface Morphism<T = any, U = any> {
  (value: T): U
}

export type Primitive = string | number | boolean | symbol | bigint | null | undefined

export interface ToString {
  toString(): string
}

export function staticImplements<T>() {
  return <U extends T>(constructor: U) => { constructor }
}

export function isPlainObject(value: any): boolean {
  return Object.getPrototypeOf(value) === Object.prototype
}

export function isFunction(value: any): value is Function {
  return value != null && typeof value === 'function'
}

export function isIterable(value: any): value is Iterable<any> {
  return isFunction(value[Symbol.iterator])
}

export function isString(value: any): value is string {
  return typeof value === 'string'
}

export function isNumber(value: any): value is number {
  return !isNaN(value)
}

export type Constructor<T = any> = new (...args: any[]) => T

export function orElse(thisArg: any, a: Nullable<Function>, b: Function) {
  return function(...args: any) {
    const fn = a != null ? a : b
    return fn.apply(thisArg, args)
  }
}

export function ifNull(thisArg: any, b: Function, ...args: any) {
  return function(a: Nullable<Function>) {
    return orElse(thisArg, a, b).apply(thisArg, args)
  }
}

function applyMixins(derivedCtor: any, constructors: any[]) {
  constructors.forEach((baseCtor) => {
    Object.getOwnPropertyNames(baseCtor.prototype).forEach((name) => {
      Object.defineProperty(
        derivedCtor.prototype,
        name,
        Object.getOwnPropertyDescriptor(baseCtor.prototype, name) ||
        Object.create(null)
      )
    })
  })
}

export function mixin<U = any>(impl: Function) {
  return function <TBase extends Constructor<U>>(constructor: TBase) {
    applyMixins(constructor, [impl])
    return constructor
  }
}

type AnyFunc = (...arg: any) => any

type PipeArgs<F extends AnyFunc[], Acc extends AnyFunc[] = []> = F extends [
  (...args: infer A) => infer B
]
  ? [...Acc, (...args: A) => B]
  : F extends [(...args: infer A) => any, ...infer Tail]
  ? Tail extends [(arg: infer B) => any, ...any[]]
  ? PipeArgs<Tail, [...Acc, (...args: A) => B]>
  : Acc
  : Acc

type LastFnReturnType<F extends Array<AnyFunc>, Else = never> = F extends [
  ...any[],
  (...arg: any) => infer R
] ? R : Else

export function pipe<FirstFn extends AnyFunc, F extends AnyFunc[]>(
  arg: Parameters<FirstFn>[0],
  firstFn: FirstFn,
  ...fns: PipeArgs<F> extends F ? F : PipeArgs<F>
): LastFnReturnType<F, ReturnType<FirstFn>> {
  return (fns as AnyFunc[]).reduce((acc, fn) => fn(acc), firstFn(arg))
}

